function [V,d] = invariances(H,f, xp),
% INVARIANCES computes the invariances of a quadratic form.
%   [V,D] = INVARIANCES(H,F, XP) computes the invariances V of the
%   quadratic form 0.5*x'*H*x + f'*x + c in the optimal stimulus XP.
%   Return in V the direction of the invariances, sorted by increasing
%   magnitude of the second derivative and in D the corresponding second
%   derivative.
%
%   Reference: Pietro Berkes and Laurenz Wiskott (2005) On the analysis
%   and interpretation of quadratic forms as receptive fields.

  % input dimension
  dim = size(H,1);

  % radius of the sphere
  r = norm(xp);
  % constant term in the second derivative
  cte = -r^(-2)*( xp'*H*xp + f'*xp );

  %% get a basis of the tangential plane in xp
  one = eye(dim);
  basis = [xp, one(:,2:dim)];
  B = gram_schmidt(basis);
  B = B(:,2:dim);
  % restrict H to the tangential plane 
  Ht = B'*H*B;
  [V,D] = eig(Ht);
  % second derivative in the direction of the eigenvectors
  d = diag(D)+cte;
  [tmp,idx] = sort(abs(d));
  % sort by increasing absolute value of the second derivative
  d = d(idx);
  V = V(:,idx);
  % project the eigenvectors back in R^N
  V = B*V;
