function x = optimal_stimulus(H,f, r, eps),
% OPTIMAL_STIMULUS computes the optimal stimulus of a quadratic form.
%   X = OPTIMAL_STIMULUS(H,F, R, EPS) computes the optimal excitatory stimulus
%   of the quadratic form 0.5*x'*H*x + f'*x + c, i.e. the input vector X
%   that maximizes the quadratic form given a fixed norm R.
%
%   EPS tolerance of norm(X) from R.
%
%   This function can be used to compute the optimal inhibitory stimulus
%   by calling it with the negative of the quadratic form -H, -F .
%
%   Reference: Pietro Berkes and Laurenz Wiskott (2006) On the analysis
%   and interpretation of quadratic forms as receptive fields.
  
  % input dimension
  dim = size(H,1);

  % compute the eigenvalues mu and eigenvectors V
  [V,D] = eig(H);
  mu = diag(D)';
  % compute the coefficients of the eigenvectors decomposition of f
  alpha = V'*f;

  % compute the range of the parameter lambda
  %  left bound for lambda
  %   added 'real' to avoid numerical problems if you maximize a
  %   ill-conditioned quadratic form
  lambda_left = max(real(mu));
  % right bound for lambda
  lambda_right = norm(f)/r + lambda_left;
  
  % search by bisection until norm(x)^2 = r^2
  r_2 = r^2;
  % norm_x_2 holds the value of norm(x)^2 at the current lambda
  norm_x_2 = 0;
  while abs(sqrt(norm_x_2)-r) > eps,
    % bisect the lambda interval
    lambda = (lambda_right-lambda_left)/2 + lambda_left;
    % compute the eigenvalues of (lambda*Id - H)^-1
    beta = (lambda-mu).^(-1);
	  
    % compute norm(x)^2 at lambda
    norm_x_2 = sum(beta'.^2.*alpha.^2);

    % update the interval limits
    if norm_x_2 > r_2,
      lambda_left = lambda;
    else
      lambda_right = lambda;
    end
  end
  
  % lambda found, compute the solution
  x = sum(repmat(beta,dim,1).*V.*repmat(alpha',dim,1),2);
